import os
import shlex
from pathlib import Path
from typing import List

import yaml

from . import exceptions, file_contents, helpers, steps
from .configs import Config


class Runtime:
    def __init__(self, config: Config, operation_id: str, **kwargs):
        self.config = config
        self.operation_id = operation_id
        self.memory = kwargs.get('memory')
        self.cpu = kwargs.get('cpu')
        self.env = kwargs.get('env')

    @property
    def command_build(self) -> str:
        return f"gcloud builds submit --pack image=gcr.io/{self.config.project}/{self.operation_id} --project={self.config.project}"

    @property
    def command_create(self) -> str:
        name = self.operation_id.replace('_', '-')

        env = ""
        if self.env is not None:
            env_vars = ",".join(self.env)
            env = f"--set-env-vars {env_vars}"
        else:
            env = "--set-env-vars ENV='production'"

        return " ".join([
            f"gcloud beta run jobs create {name}"
            f" --image='gcr.io/{self.config.project}/{self.operation_id}' --tasks 1 {env}"
            f" --max-retries 3 --region={self.config.region} --project={self.config.project}",
            "" if self.memory is None else f"--memory={self.memory}",
            "" if self.cpu is None else f"--cpu={self.cpu}",
        ])

    @property
    def command_delete(self) -> str:
        name = self.operation_id.replace('_', '-')
        return f"gcloud beta run jobs delete {name} --region='{self.config.region}' --project='{self.config.project}'"

    @property
    def command_run(self) -> str:
        name = self.operation_id.replace('_', '-')
        return f"gcloud beta run jobs execute {name} --region='{self.config.region}' --project='{self.config.project}'"

    @property
    def command_run_dev(self) -> str:
        return "python3 main.py"

    @property
    def secrethub_env(self) -> str:
        return "secrethub inject -i .env.temp -o .env"


class Job:
    def __init__(self, operation_id: str, **kwargs):
        self.operation_id = operation_id

    def runtime(self, config: Config, **kwargs) -> Runtime:
        is_debug = kwargs.get('is_debug', False)
        config_path = f'{config.base_dir}/jobs/config.yaml'
        if Path(config_path).exists():
            with open(config_path) as f:
                runtime_config = yaml.safe_load(f.read())
        else:
            runtime_config = {}

        env_path = f'{config.base_dir}/jobs/{self.operation_id}/.env'
        if Path(env_path).exists():
            with open(env_path) as f:
                env_data = f.read().split('\n')
        else:
            env_data = None

        if is_debug:
            print('ENV Data', env_data)
        return Runtime(config, self.operation_id, env=env_data, **runtime_config)


def job_list(config: Config) -> List[Job]:
    jobs_path = f'{config.base_dir}/jobs'
    if not Path(jobs_path).exists():
        Path(jobs_path).mkdir(parents=True, exist_ok=True)
    return [
        Job(x) for x in os.listdir(f'{config.base_dir}/jobs')
    ]


def job_name_clean(value: str) -> str:
    return value.replace(' ', '_').lower()


def cli_main(args, config: Config):
    if not config.load():
        return 1
    if args.action == 'init':
        manager = steps.Manager(config, [
            steps.StepCommand(
                shlex.split(
                    f'gcloud services enable run.googleapis.com --project="{config.project}"'
                )
            ),
            steps.StepCommand(
                shlex.split(
                    f'gcloud services enable cloudscheduler.googleapis.com --project="{config.project}"'
                )
            ),
        ], simulate=args.simulate)
        rc = manager.run()
        if rc != 0:
            return rc
        print(helpers.bcolors.OKGREEN + 'Init API-Gateway success' + helpers.bcolors.ENDC)
        return 0
    if args.action == 'list':
        jobs = list(job_list(config))
        for job in jobs:
            print(job.operation_id)
        return 0
    elif args.action == 'create':
        job_name = job_name_clean(args.name)
        jobs = [x for x in job_list(config) if x.operation_id == job_name]
        if len(jobs) > 0:
            print(
                helpers.bcolors.FAIL
                + f'Job {args.endpoint} already exists'
                + helpers.bcolors.ENDC
            )
            return 1

        commands = [
            steps.StepCreatDir(f'{config.base_dir}/jobs/{job_name}'),
            steps.StepCreateFile(
                f'{config.base_dir}/jobs/{job_name}/main.py',
                file_contents.job_main(),
            ),
            steps.StepCreateFile(
                f'{config.base_dir}/jobs/{job_name}/Procfile',
                file_contents.job_procfile(),
            ),
            steps.StepCreateFile(
                f'{config.base_dir}/jobs/{job_name}/requirements.in',
                """shopcloud-serverless
""",
            ),
            steps.StepCreateFile(
                f'{config.base_dir}/jobs/{job_name}/requirements.txt',
                """shopcloud-serverless
"""
            ),
            steps.StepCreateFile(
                f'{config.base_dir}/jobs/{job_name}/.env.temp',
                "ENV=production",
            ),
            # gitignore
            # cloudignore
        ]
        manager = steps.Manager(config, commands, simulate=args.simulate)
        rc = manager.run()
        if rc != 0:
            return rc
        return 0
    elif args.action == 'describe':
        job_name = job_name_clean(args.name)
        jobs = [x for x in job_list(config) if x.operation_id == job_name]
        if len(jobs) == 0:
            print(
                helpers.bcolors.FAIL
                + f'Job {job_name} not found'
                + helpers.bcolors.ENDC
            )
            return 1
        print('Operation ID:', jobs[0].operation_id)
        return 0
    elif args.action == 'deploy':
        job_name = job_name_clean(args.name)
        jobs = [x for x in job_list(config) if x.operation_id == job_name]
        if len(jobs) == 0:
            print(
                helpers.bcolors.FAIL
                + f'Job {job_name} not found'
                + helpers.bcolors.ENDC
            )
            return 1

        print('# Deploy')
        job = jobs[0]
        runtime = job.runtime(config, is_debug=args.debug)
        commands = [x for x in [
            # delete job if exists?
            steps.StepCommand(
                shlex.split(runtime.command_build),
                work_dir=f'{config.base_dir}/jobs/{job_name}',
            ),
            None if not Path(f'{config.base_dir}/jobs/{job_name}/.env.temp').exists() else steps.StepCommand(
                shlex.split(runtime.secrethub_env),
                work_dir=f'{config.base_dir}/jobs/{job_name}',
            ),
            steps.StepCommand(
                shlex.split(runtime.command_delete),
                work_dir=f'{config.base_dir}/jobs/{job_name}',
                can_fail=True,
                retry=3,
            ),
            steps.StepCommand(
                shlex.split(runtime.command_create),
                work_dir=f'{config.base_dir}/jobs/{job_name}',
                retry=3,
            ),
        ] if x is not None]
        manager = steps.Manager(config, commands, simulate=args.simulate)
        rc = manager.run()
        if rc != 0:
            return rc

        return 0
    elif args.action == 'run':
        job_name = job_name_clean(args.name)
        jobs = [x for x in job_list(config) if x.operation_id == job_name]
        if len(jobs) == 0:
            print(
                helpers.bcolors.FAIL
                + f'Job {job_name} not found'
                + helpers.bcolors.ENDC
            )
            return 1

        print('# Deploy')
        job = jobs[0]
        runtime = job.runtime(config)
        commands = [
            steps.StepCommand(
                shlex.split(runtime.command_run),
                work_dir=f'{config.base_dir}/jobs/{job_name}',
            ),
        ]
        manager = steps.Manager(config, commands, simulate=args.simulate)
        rc = manager.run()
        if rc != 0:
            return rc
        return 0
    elif args.action == 'run-dev':
        job_name = job_name_clean(args.name)
        jobs = [x for x in job_list(config) if x.operation_id == job_name]
        if len(jobs) == 0:
            print(
                helpers.bcolors.FAIL
                + f'Job {job_name} not found'
                + helpers.bcolors.ENDC
            )
            return 1
        job = jobs[0]
        runtime = job.runtime(config)
        commands = [x for x in [
            None if Path(f'{config.base_dir}/jobs/{job_name}/.env').exists() else steps.StepCommand(
                shlex.split(runtime.secrethub_env),
                work_dir=f'{config.base_dir}/jobs/{job_name}',
            ),
            steps.StepCommand(
                shlex.split(runtime.command_run_dev),
                work_dir=f'{config.base_dir}/jobs/{job_name}',
            ),
        ] if x is not None]
        manager = steps.Manager(config, commands, simulate=args.simulate)
        rc = manager.run()
        if rc != 0:
            return rc
        return 0



import shutil
import subprocess
from pathlib import Path
from typing import List

from . import exceptions, helpers
from .configs import Config


class Step:
    def __init__(self):
        pass

    def run(self, config: Config, simulate: bool = False):
        raise NotImplementedError()


class StepCreatDir(Step):
    def __init__(self, path: str):
        self.path = path

    def run(self, config: Config, simulate: bool = False):
        print(f'+ Creating {self.path} directory')
        if not Path(self.path).exists():
            Path(self.path).mkdir(parents=True, exist_ok=True)


class StepDeleteDir(Step):
    def __init__(self, path: str):
        self.path = path

    def run(self, config: Config, simulate: bool = False):
        if not Path(self.path).exists():
            return None
        print(f'+ Deleting {self.path} directory')
        shutil.rmtree(self.path)


class StepCopyFileContent(Step):
    def __init__(self, file_a_path: str, file_b_path: str):
        self.file_a_path = file_a_path
        self.file_b_path = file_b_path

    def run(self, config: Config, simulate: bool = False):
        print(f'+ Copy {self.file_a_path} to {self.file_b_path}')
        with open(self.file_a_path) as fr:
            with open(self.file_b_path, "w") as fw:
                fw.write(fr.read())


class StepAppendFileContent(Step):
    def __init__(self, file_path: str, content: List[str]):
        self.file_path = file_path
        self.content = content

    def run(self, config: Config, simulate: bool = False):
        print(f'+ Append to {self.file_path}')
        with open(self.file_path, "a") as f:
            f.write("\n".join(self.content))


class StepCreateFile(Step):
    def __init__(self, file_path: str, content: str):
        self.file_path = file_path
        self.content = content

    def run(self, config: Config, simulate: bool = False):
        print(f'+ Append to {self.file_path}')
        with open(self.file_path, "w") as f:
            f.write(self.content)


class StepCopyDir(Step):
    def __init__(self, path_a: str, path_b: str):
        self.path_a = path_a
        self.path_b = path_b

    def run(self, config: Config, simulate: bool = False):
        print(f'+ Copy {self.path_a} to {self.path_b}')
        shutil.copytree(self.path_a, self.path_b)


class StepCommand(Step):
    def __init__(self, command: List[str], **kwargs):
        self.command = command
        self.work_dir = kwargs.get('work_dir')
        self.can_fail = kwargs.get('can_fail', False)
        self.retry = kwargs.get('retry', 0)

    def run(self, config: Config, simulate: bool = False):
        print('+ Run command:')
        print(" ".join(self.command))
        if not simulate:
            while True:
                count = 0
                p = subprocess.run(self.command, stdout=subprocess.PIPE, cwd=self.work_dir)
                if not self.can_fail:
                    if p.returncode != 0:
                        if count < self.retry:
                            count += 1
                            continue
                        raise exceptions.CommandError('command not success')
                break


class StepFunction(Step):
    def __init__(self, name: str, function):
        self.name = name
        self.function = function

    def run(self, config: Config, simulate: bool = False):
        print(f'+ Run {self.name}')
        self.function(config)


class Manager:
    def __init__(self, config: Config, steps: list, **kwargs):
        self.config = config
        self.steps = steps
        self.simulate = kwargs.get('simulate', False)

    def run(self) -> int:
        if self.simulate:
            print(helpers.bcolors.WARNING + '+ Simulate' + helpers.bcolors.ENDC)
        for step in self.steps:
            try:
                step.run(self.config, self.simulate)
            except Exception as e:
                print(helpers.bcolors.FAIL + f'ERROR: {e}' + helpers.bcolors.ENDC)
                return 1
        return 0



import shlex

from . import exceptions, helpers, steps
from .configs import Config


def cli_main(args, config: Config) -> int:
    if not config.load():
        return 1
    if args.action == 'init':
        manager = steps.Manager(config, [
            steps.StepCommand(
                shlex.split(
                    f'gcloud services enable apigateway.googleapis.com --project="{config.project}"'
                )
            ),
            steps.StepCommand(
                shlex.split(
                    f'gcloud services enable servicemanagement.googleapis.com --project="{config.project}"'
                )
            ),
            steps.StepCommand(
                shlex.split(
                    f'gcloud services enable servicecontrol.googleapis.com --project="{config.project}"'
                )
            ),
            steps.StepCommand(
                shlex.split(
                    f'gcloud services enable cloudbuild.googleapis.com --project="{config.project}"'
                )
            ),
            steps.StepCommand(
                shlex.split(
                    f'gcloud services enable cloudfunctions.googleapis.com --project="{config.project}"'
                )
            ),
        ], simulate=args.simulate)
        rc = manager.run()
        if rc != 0:
            return rc
        print(helpers.bcolors.OKGREEN + 'Init API-Gateway success' + helpers.bcolors.ENDC)
        return 0
    elif args.action == 'deploy':
        config.endpoint_deploy_version = config.endpoint_deploy_version + 1
        api_id = config.endpoint_api_title.lower().replace(' ', '-').replace('_', '-')
        api_config_id = f"v{config.endpoint_deploy_version}"

        if config.endpoint_deploy_version == 1:
            if args.debug:
                print(f'Creating API Gateway {api_id}')
            commands = [
                steps.StepCommand(
                    shlex.split(
                        f'gcloud api-gateway api-configs create {api_config_id} --api={api_id} --openapi-spec=api.yaml --project="{config.project}" --backend-auth-service-account="{config.project}@appspot.gserviceaccount.com"'
                    ),
                    work_dir=f'{config.base_dir}',
                ),
                steps.StepCommand(
                    shlex.split(
                        f'gcloud api-gateway gateways create {api_id} --api={api_id} --api-config={api_config_id} --location={config.region} --project="{config.project}"'
                    ),
                    work_dir=f'{config.base_dir}',
                )
            ]
        else:
            if args.debug:
                print(f'Updating API Gateway {api_id}')
            commands = [
                steps.StepCommand(
                    shlex.split(
                        f'gcloud api-gateway api-configs create {api_config_id} --api={api_id} --openapi-spec=api.yaml --project="{config.project}" --backend-auth-service-account="{config.project}@appspot.gserviceaccount.com"'
                    ),
                    work_dir=f'{config.base_dir}',
                ),
                steps.StepCommand(
                    shlex.split(
                        f'gcloud api-gateway gateways update {api_id} --api={api_id} --api-config={api_config_id} --location={config.region} --project="{config.project}"'
                    ),
                    work_dir=f'{config.base_dir}',
                )
            ]

        manager = steps.Manager(config, commands, simulate=args.simulate)
        rc = manager.run()
        if rc != 0:
            return rc
        print(helpers.bcolors.OKGREEN + 'Deploy API-Gatewway success' + helpers.bcolors.ENDC)
        config.save()
        return 0



import os
from pathlib import Path


class Environment:
    def __init__(self):
        self._data = self._read_env()

    def _read_env(self):
        env_data = {}
        if Path('.env').exists():
            with open('.env') as f:
                env_data = f.read().split('\n')
                env_data = {x.split('=')[0]: x.split('=')[1] for x in env_data if x}
        return env_data

    def get(self, key: str):
        return self._data.get(key, os.environ.get(key))



import shlex
from pathlib import Path
from typing import List

import yaml

from . import file_contents, helpers, steps
from .configs import Config


class Runtime:
    def __init__(self, config: Config, operation_id: str, **kwargs):
        self.config = config
        self.operation_id = operation_id
        self.entry_point = kwargs.get('entry_point', 'main')
        self.runtime = kwargs.get('runtime', 'python311')
        self.memory = kwargs.get('memory', '256MB')
        self.dependencies = kwargs.get('dependencies', [])
        self.trigger = kwargs.get('trigger', 'http')
        if self.trigger == 'http':
            self.trigger = '--trigger-http'
        else:
            self.trigger = f"--trigger-topic='{self.trigger}'"

    @property
    def command(self) -> str:
        return f'gcloud functions deploy --project="{self.config.project}" {self.operation_id} {self.trigger} --env-vars-file env.yaml --runtime {self.runtime} --entry-point="{self.entry_point}" --memory={self.memory} --region="{self.config.region}"'


class Endpoint:
    def __init__(self, operation_id: str, **kwargs):
        self.operation_id = operation_id
        self.summary = kwargs.get('summary', '')
        self.config = kwargs.get('config', None)
        self.runtime_config = None

    def _load_runtime_config(self, config: Config):
        if self.runtime_config is not None:
            return self.runtime_config
        with open(f'{config.base_dir}/endpoints/{self.operation_id}.yaml') as f:
            self.runtime_config = yaml.safe_load(f.read())
        return self.runtime_config

    def runtime(self, config: Config) -> Runtime:
        runtime_config = self._load_runtime_config(config)
        return Runtime(config, self.operation_id, **runtime_config)

    def _env_secrets(self, config: Config, **kwargs) -> List[dict]:
        runtime_config = self._load_runtime_config(config)
        secrets = {x.get('name'): "{{ "+x.get('value')+" }}" for x in runtime_config.get('secrets', [])}
        return secrets

    def env(self, config: Config, **kwargs) -> str:
        is_simulate = kwargs.get('is_simulate', False)
        env = {
            **self._env_secrets(config, is_simulate=is_simulate),
            **{
                'ENV': 'production',
            }
        }
        env = yaml.dump(env, default_flow_style=False)
        env = env.replace("'{{", "{{").replace("}}'", '}}')
        return env


def endpoint_load(config: Config) -> dict:
    with open(f"{config.base_dir}/api.yaml") as f:
        data = yaml.safe_load(f)
    return data


def endpoint_list(config: Config) -> List[Endpoint]:
    data = endpoint_load(config)
    endpoints = list(data.get('paths', {}).values())
    endpoints = [
        [Endpoint(y.get('operationId'), summary=y.get('summary'), config=y) for y in x.values()]
        for x in endpoints
    ]
    return [item for sublist in endpoints for item in sublist]


def endpoint_create(base_dir: str, name: str):
    endpoint_path = 'endpoints'
    if not Path(base_dir, 'src', 'services', f'{name}.py').exists():
        with open(f"{base_dir}/{endpoint_path}/src/services/{name}.py", "w") as file:
            file.write(file_contents.endpoint_lib(name))
    if not Path(base_dir, endpoint_path).exists():
        Path(base_dir, endpoint_path).mkdir()
    if not Path(base_dir, f'{endpoint_path}/{name}.py').exists():
        with open(f"{base_dir}/{endpoint_path}/{name}.py", "w") as file:
            file.write(file_contents.endpoint_main(name))
    if not Path(base_dir, f'{endpoint_path}/{name}.yaml').exists():
        with open(f"{base_dir}/{endpoint_path}/{name}.yaml", "w") as file:
            file.write(file_contents.endpoint_config(name))


def init_file_structure(config: Config):
    base_dir = config.base_dir
    if base_dir != '.':
        if not Path(base_dir).exists():
            Path(base_dir).mkdir()
    if not Path(base_dir, 'endpoints').exists():
        Path(base_dir, 'endpoints').mkdir()
    if not Path(base_dir, 'endpoints', 'src').exists():
        Path(base_dir, 'endpoints', 'src').mkdir()
        with open(f"{base_dir}/endpoints/src/__init__.py", "w") as file:
            file.write("""from . import services
""")
        Path(base_dir, 'endpoints', 'src', 'services').mkdir()
        with open(f"{base_dir}/endpoints/src/services/__init__.py", "w") as file:
            file.write("""from . import docs
from . import hello_world
""")
        endpoint_create(base_dir, 'hello_world')
        endpoint_create(base_dir, 'docs')

    if not Path(base_dir, 'api.yaml').exists():
        with open(f"{base_dir}/api.yaml", "w") as file:
            file.write(file_contents.api(
                title=config.endpoint_api_title,
                description=config.endpoint_api_description,
                project=config.project,
                region=config.region,
            ))
    if not Path(base_dir, 'requirements.txt').exists():
        with open(f"{base_dir}/requirements.txt", "w") as file:
            file.write(file_contents.requirements())
    if not Path(base_dir, 'requirements.in').exists():
        with open(f"{base_dir}/requirements.in", "w") as file:
            file.write(file_contents.requirements())


def cli_main(args, config: Config) -> int:
    if not config.load():
        return 1

    if args.action == 'init':
        manager = steps.Manager(config, [
            steps.StepFunction(
                'init-file-structure',
                lambda x: init_file_structure(config)
            ),
            steps.StepCommand(
                shlex.split(
                    f'gcloud services enable cloudfunctions.googleapis.com --project="{config.project}"'
                )
            ),
            steps.StepCommand(
                shlex.split(
                    f'gcloud services enable cloudbuild.googleapis.com --project="{config.project}"'
                )
            ),
        ], simulate=args.simulate)
        rc = manager.run()
        if rc != 0:
            return rc
        print(helpers.bcolors.OKGREEN + 'Init API-Endpoint success' + helpers.bcolors.ENDC)
        return 0
    elif args.action == 'list':
        endpoints = endpoint_list(config)
        for endpoint in endpoints:
            print(endpoint.operation_id, '-', endpoint.summary)
        return 0
    elif args.action == 'create':
        endpoints = [x for x in endpoint_list(config) if x.operation_id == args.endpoint]
        if len(endpoints) > 0:
            print(
                helpers.bcolors.FAIL
                + f'Endpoint {args.endpoint} already exists'
                + helpers.bcolors.ENDC
            )
            return 1
        endpoint_create(config.base_dir, args.endpoint)
        print(helpers.bcolors.OKGREEN + 'Endpoint created' + helpers.bcolors.ENDC)
    elif args.action == 'describe':
        endpoints = [x for x in endpoint_list(config) if x.operation_id == args.endpoint]
        if len(endpoints) == 0:
            print(
                helpers.bcolors.FAIL
                + f'Endpoint {args.endpoint} not found'
                + helpers.bcolors.ENDC
            )
            return 1
        print('Operation ID:', endpoints[0].operation_id)
        print('Summary:', endpoints[0].summary)
        return 0
    elif args.action == 'deploy':
        endpoints = [x for x in endpoint_list(config) if x.operation_id == args.endpoint]
        if len(endpoints) == 0:
            print(
                helpers.bcolors.FAIL
                + f'Endpoint {args.endpoint} not found'
                + helpers.bcolors.ENDC
            )
            return 1

        print('# Deploy')
        endpoint = endpoints[0]
        runtime = endpoint.runtime(config)
        manager = steps.Manager(config, [
            steps.StepDeleteDir(f'{config.base_dir}/build/endpoints/{endpoint.operation_id}'),
            steps.StepCreatDir(f'{config.base_dir}/build/endpoints/{endpoint.operation_id}'),
            steps.StepCopyFileContent(
                f'{config.base_dir}/requirements.txt',
                f'{config.base_dir}/build/endpoints/{args.endpoint}/requirements.txt'
            ),
            steps.StepAppendFileContent(
                f'{config.base_dir}/build/endpoints/{args.endpoint}/requirements.txt',
                runtime.dependencies
            ),
            steps.StepCopyFileContent(
                f'{config.base_dir}/endpoints/{endpoint.operation_id}.py',
                f'{config.base_dir}/build/endpoints/{endpoint.operation_id}/main.py'
            ),
            steps.StepCopyFileContent(
                f'{config.base_dir}/api.yaml',
                f'{config.base_dir}/build/endpoints/{endpoint.operation_id}/api.yaml'
            ),
            steps.StepCopyDir(
                f'{config.base_dir}/endpoints/src',
                f'{config.base_dir}/build/endpoints/{endpoint.operation_id}/src'
            ),
            steps.StepCreateFile(
                f'{config.base_dir}/build/endpoints/{endpoint.operation_id}/env.temp.yaml',
                endpoint.env(config, is_simulate=args.simulate)
            ),
            steps.StepCommand(
                shlex.split("secrethub inject -i env.temp.yaml -o env.yaml"),
                work_dir=f'{config.base_dir}/build/endpoints/{endpoint.operation_id}',
            ),
            steps.StepCommand(
                shlex.split(runtime.command),
                work_dir=f'{config.base_dir}/build/endpoints/{endpoint.operation_id}',
            ),
        ], simulate=args.simulate)
        rc = manager.run()
        if rc != 0:
            return rc
        print(helpers.bcolors.OKGREEN + f'Deployed endpoint {endpoint.operation_id} success' + helpers.bcolors.ENDC)
        return 0



from . import endpoints, gateways, helpers, jobs
from .configs import Config


def main(args) -> int:
    if not hasattr(args, 'which'):
        print(
            helpers.bcolors.FAIL
            + 'Can not parse action use --help'
            + helpers.bcolors.ENDC
        )
        return 1

    if hasattr(args, 'debug') and args.debug:
        print(args)

    config = Config(config_filename=args.config)

    if args.which == 'init':
        config.base_dir = args.base_dir or helpers.ask_for('Base directory', '.')

        # Cloud
        config.cloud = args.cloud or helpers.ask_for('Cloud', 'gcp')
        config.project = args.project or helpers.ask_for('Cloud project')
        config.region = args.region or helpers.ask_for('Cloud region', 'europe-west1')

        # Endpoints
        config.endpoint_api_title = args.api_title or helpers.ask_for('Endpoint API title', 'My API')
        config.endpoint_api_description = args.api_description or helpers.ask_for('Endpoint API description', 'My API description')

        config.save()
        print(helpers.bcolors.OKGREEN + 'Init config success' + helpers.bcolors.ENDC)
        return 0
    elif args.which == 'jobs':
        return jobs.cli_main(args, config)
    elif args.which == 'gateway':
        return gateways.cli_main(args, config)
    elif args.which == 'endpoints':
        return endpoints.cli_main(args, config)
    elif args.which == 'memorystore':
        if not config.load():
            return 1
        print('# Init Memory Store')
        print(f'- go to https://console.cloud.google.com/memorystore/&project={config.project} and create a redis instance')
        print('- set network as default')
        print('# Init Serverloser VPC-Zugriff')
        print(f'- go to https://console.cloud.google.com/networking/connectors/list?project={config.project}')
        print(f'- create a connector with name `serverless-vpc-connector` and region `{config.region}`')
        print('  - with network 10.8.0.0/28')
        print('')
        print('to use it in the gcloud functions use the --vpc-connector option')



def api(**kwargs):
    api_title = kwargs.get('api_title', 'My API')
    api_description = kwargs.get('api_description', 'My API Description')
    project = kwargs.get('project')
    region = kwargs.get('region')
    return f"""swagger: '2.0'
info:
  title: {api_title}
  description: {api_description}
  version: 0.0.0
schemes:
  - https
produces:
  - application/json
securityDefinitions:
  api_key:
    type: "apiKey"
    name: "key"
    in: "query"
paths:
  /hello-world:
    get:
      tags:
      - "Main"
      summary: Hello World
      operationId: hello_world
      consumes:
      - "application/json"
      x-google-backend:
        address: https://{region}-{project}.cloudfunctions.net/hello_world
      security:
      - api_key: []
      responses:
        "200":
          description: A successful response
          schema:
            $ref: "#/definitions/HelloWorldResponse"
  /docs:
    get:
      tags:
      - "Documentation"
      summary: Documentation
      operationId: docs
      x-google-backend:
        address: https://{region}-{project}.cloudfunctions.net/docs
      parameters:
      - in: "query"
        type: "string"
        name: "content"
        description: "Specify the content"
        enum: ["docs", "openapi.json"]
        default: "docs"
      responses:
        "200":
          description: A successful response
definitions:
  HelloWorld:
    type: "object"
    required:
      - value
    properties:
      value:
        type: "string"
  HelloWorldResponse:
    type: "object"
    required:
      - value
    properties:
      status:
        type: "string"

"""


def endpoint_main(name: str, **kwargs):
    if name == 'docs':
        return """import yaml
import os
import importlib
import functions_framework
from flask import jsonify

try:
    import src
except Exception:
    src = importlib.import_module('src')


@functions_framework.http
def main(request):
    if request.args.get('content') == 'openapi.json':
        file = "api.yaml"
        if os.environ.get('ENV') != 'production':
            file = "./../api.yaml"
        with open(file) as f:
            data = yaml.safe_load(f.read())
        return jsonify(data)

    return src.services.docs.template_docs()

"""

    return """
import importlib
import functions_framework
from flask import jsonify

try:
    import src
except Exception:
    src = importlib.import_module('src')


@functions_framework.http
def main(request):
    return jsonify({'status': src.services.#SERVICE#.get_hello_world()})

""".replace('#SERVICE#', name)


def endpoint_config(name: str, **kwargs):
    return """trigger: http
"""


def endpoint_lib(name: str, **kwargs):
    if name == 'docs':
        return """def template_docs():
    return \"""<!DOCTYPE html>
<html>
<head>
<link type="text/css" rel="stylesheet" href="https://cdn.jsdelivr.net/npm/swagger-ui-dist@4/swagger-ui.css">
<link rel="shortcut icon" href="https://fastapi.tiangolo.com/img/favicon.png">
<title>Sage-Gateway - Swagger UI</title>
</head>
<body>
<div id="swagger-ui">
</div>
<script src="https://cdn.jsdelivr.net/npm/swagger-ui-dist@4/swagger-ui-bundle.js"></script>
<!-- `SwaggerUIBundle` is now available on the page -->
<script>
const ui = SwaggerUIBundle({
    url: '/docs?content=openapi.json',
"dom_id": "#swagger-ui",
"layout": "BaseLayout",
"deepLinking": true,
"showExtensions": true,
"showCommonExtensions": true,
oauth2RedirectUrl: window.location.origin + '/docs/oauth2-redirect',
presets: [
    SwaggerUIBundle.presets.apis,
    SwaggerUIBundle.SwaggerUIStandalonePreset
    ],
})
</script>
</body>
</html>
\"""

"""

    return """def get_hello_world():
    return "Hello World"

    """


def requirements():
    return """functions-framework
pyyaml
"""


def job_main():
    return """from shopcloud_serverless import Environment

if __name__ == "__main__":
    env = Environment()
    print(env.get('ENV'))
    print("Hello World")

"""


def job_procfile():
    return """web: python3 main.py"""



import argparse
import sys

from . import cli

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Serverless',
        prog='shopcloud-serverless'
    )

    subparsers = parser.add_subparsers(help='commands', title='commands')
    parser.add_argument('--debug', '-d', help='Debug', action='store_true')
    parser.add_argument('--simulate', '-s', help='Simulate the process', action='store_true')
    parser.add_argument('--secrethub-token', help='Secrethub-Token', type=str)
    parser.add_argument('--config', help='The Config Filename', type=str)

    parser_init = subparsers.add_parser('init', help='init the eventbus')
    parser_init.add_argument('--base-dir', help='Base directory', type=str)
    parser_init.add_argument('--api-title', help='API title', type=str)
    parser_init.add_argument('--api-description', help='API description', type=str)
    parser_init.add_argument('--cloud', help='Cloud type', type=str, default='gcp')
    parser_init.add_argument('--project', help='Cloud project id', type=str)
    parser_init.add_argument('--region', help='Cloud region', type=str)
    parser_init.set_defaults(which='init')

    parser_endpoint = subparsers.add_parser('endpoints', help='endpoints')
    parser_endpoint.add_argument(
        'action',
        const='generate',
        nargs='?',
        choices=['init', 'list', 'create', 'describe', 'deploy']
    )
    parser_endpoint.add_argument('endpoint', const='generate', nargs='?')
    parser_endpoint.set_defaults(which='endpoints')

    parser_gateway = subparsers.add_parser('gateway', help='api gateway')
    parser_gateway.add_argument(
        'action',
        const='generate',
        nargs='?',
        choices=['init', 'deploy']
    )
    parser_gateway.set_defaults(which='gateway')

    parser_jobs = subparsers.add_parser('jobs', help='jobs')
    parser_jobs.add_argument(
        'action',
        const='generate',
        nargs='?',
        choices=['init', 'list', 'create', 'describe', 'deploy', 'run', 'run-dev']
    )
    parser_jobs.add_argument('name', const='generate', nargs='?')
    parser_jobs.set_defaults(which='jobs')

    parser_memorystore = subparsers.add_parser('memorystore', help='generate a redis memorystor')
    parser_memorystore.add_argument(
        'action',
        const='generate',
        nargs='?',
        choices=['init']
    )
    parser_memorystore.set_defaults(which='memorystore')

    args = parser.parse_args()
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    rc = cli.main(args)
    if rc != 0:
        sys.exit(rc)



from pathlib import Path

import yaml

from . import helpers


class Config:
    VERSION = 'V4'

    def __init__(self, **kwargs):
        self.filename = kwargs.get('config_filename')
        if self.filename is None:
            self.filename = 'serverless.yaml'
        self.version = None
        self.base_dir = None
        self.cloud_type = None
        self.cloud_project = None
        self.region = None
        self.endpoint_api_title = None
        self.endpoint_api_description = None
        self.endpoint_deploy_version = 0

    def load(self) -> bool:
        if not Path(self.filename).exists():
            print(
                helpers.bcolors.FAIL
                + 'Config file not found. Please run `init` first.'
                + helpers.bcolors.ENDC
            )
            return False

        with open(self.filename) as f:
            data = yaml.safe_load(f)
            self.version = data.get('version', '')
            self.base_dir = data.get('base_dir', '.')
            self.cloud = data.get('cloud')
            self.project = data.get('project')
            self.region = data.get('region')
            self.endpoint_api_title = data.get('endpoints', {}).get('api_title')
            self.endpoint_api_description = data.get('endpoints', {}).get('api_description')
            self.endpoint_deploy_version = int(data.get('endpoints', {}).get('deploy_version', 0))

        if str(self.version).strip() != self.VERSION:
            print(
                helpers.bcolors.FAIL
                + 'Config file is not compatible with this version. Please run `init` again.'
                + helpers.bcolors.ENDC
            )
            return False

        return True

    def save(self):
        with open(self.filename, 'w') as f:
            yaml.dump(self.dict(), f)

    def dict(self):
        return {
            'version': self.VERSION,
            'base_dir': self.base_dir or '.',
            'cloud': self.cloud,
            'project': self.project,
            'region': self.region,
            'endpoints': {
                'api_title': self.endpoint_api_title,
                'api_description': self.endpoint_api_description,
                'deploy_version': self.endpoint_deploy_version,
            },
        }



from setuptools import find_packages, setup

with open('README.md') as readme_file:
    README = readme_file.read()

setup_args = {
    "name": 'shopcloud-serverless',
    "version": '4.8.0',
    "description": 'CLI tool for creating a serverles api with API Gateway and Cloud Functions',
    "long_description_content_type": "text/markdown",
    "long_description": README,
    "license": 'MIT',
    "packages": find_packages(),
    "author": 'Konstantin Stoldt',
    "author_email": 'konstantin.stoldt@talk-point.de',
    "keywords": ['CLI'],
    "url": 'https://github.com/Talk-Point/shopcloud-serverless-cli',
    "scripts": ['./scripts/serverless'],
}

install_requires = [
    'pyyaml',
    'shopcloud-secrethub>=2.12.0',
]

if __name__ == '__main__':
    setup(**setup_args, install_requires=install_requires)



from typing import Optional


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def ask_for(variable: str, default: Optional[str] = None) -> str:
    d = '' if default is None else f' ({default})'
    v = input(f'{variable}{d}:') \
        .replace('\n', ' ') \
        .replace('\r', '') \
        .strip()
    if v == '' and default is not None:
        return default
    return v


def ask_for_yes_no(question: str):
    v = input(f'{question} (y/n):') \
        .replace('\n', ' ') \
        .replace('\r', '') \
        .strip() \
        .lower()
    return v in ['y', 'yes', 'yeahh']


def ask_for_enter():
    input('Press enter to continue...')


def fetch_secret(hub, name: str, **kwargs):
    if kwargs.get('simulate', False):
        return 'secret'
    return hub.read(name)
